import yaml, os, re
from datetime import datetime
import numpy as np
import pandas as pd
from astropy.io import fits
from astropy.coordinates import SkyCoord

from .config import  __version__, which_focalplane
from .utils import Logger
from .config import config, iso_time


default_output_dir = config['output']
log_level = config['log_level']
header_check = config['check_fits_header']

def set_up_logger(log_dir):
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    return Logger(log_dir+'/cpism_pack.log', log_level).logger

log = set_up_logger(config['log_dir'])

def check_and_update_fits_header(header):
    """
    Check the header keywords and update the description according to the data model.

    Parameters
    -----------
    header: astropy.io.fits.header.Header
        The header to be checked.

    Returns
    --------
    None
    """

    # model_file = f"{cpism_refdata}/io/image_header.json"
    # if hdu == 'primary':
    #     model_file = f"{cpism_refdata}/io/primary_header.json"
    model_file = config['datamodel']
    with open(model_file, 'r', encoding='utf-8') as fid:
        data_model = yaml.load(fid, Loader=yaml.FullLoader)

    if 'FILETYPE' in header.keys():
        header_model =data_model['HDU0']
        hdu = 'hdu0'
    else:
        header_model = data_model['HDU1']
        hdu = 'hdu1'
    dm_comment = {}

    def print_warning(info):
        if header_check:
            log.warning(info)

    # check existance and format of keyword in fits header
    for _, content in header_model.items():
        comment = content['comment']
        keyword = content['key']
        dtype = content['dtype']
        if pd.isnull(comment):
            comment = ''

        if len(comment) > 47:
            # comment = comment[:46]
            print_warning(
                f"Keyword {keyword} has a comment longer than 47 characters.")

        dm_comment[keyword] = comment

        if keyword not in header.keys():
            print_warning(f"Keyword {keyword} not found in header.")

        elif not pd.isnull(dtype):
            value = header[keyword]
            # check the type of the value, I for int, R for float, C for str
            if isinstance(value, str):
                key_type = 'str'
            elif isinstance(value, float):
                key_type = 'flo'
            elif isinstance(value, bool):
                key_type = 'boo'
            elif isinstance(value, int):
                key_type = 'int'
            elif isinstance(value, type(header['COMMENT'])):
                key_type = 'str'
            else:
                key_type = 'ukn'

            if key_type != dtype[0:3]:
                print_warning(
                    f"Keyword {keyword} has wrong type in [{hdu}]. {dtype} expected, {key_type} found.")

    # check if there are extral keyword in header, and update the comment
    for keyword in header.keys():
        # print(keyword)
        if keyword not in dm_comment.keys():
            print_warning(
                f"Keyword {keyword} not found in the [{hdu}] data model.")
        elif keyword != 'COMMENT': # comment keyword is not allowed to be updated
            header[keyword] = (header[keyword], dm_comment[keyword])

    return header


def obsid_parser(
        obsid: int):
    """
    Parse the obsid to get the obstype.

    Parameters
    ----------
    obsid: str
        The obsid of the observation.
        Obsid must be 11 digits and start with 5 for CPIC.

    Returns
    -------
    str
        The obstype of the observation.

    Raises
    ------
    ValueError
        If the obsid is not 11 digits or does not start with 5.
    """
    obsid = str(obsid)
    if len(obsid) != 11:
        raise ValueError('Obsid must be 11 digits.')

    if obsid[0] != '4':
        raise ValueError('Obsid must start with 4 for CPIC')

    obstype_dict = {
        '20': 'BIAS',
        '21': 'DARK',
        '22': 'FLAT',
        '23': 'BKG',
        '24': 'LASER',
        '01': 'SCI',
        '02': 'DSF',
        '10': 'CALS',
    }
    obstype = obstype_dict.get(obsid[1:3], 'DEFT')
    return obstype


def datetime_obj_to_mjd(time_obj):
    """
    transfer datetime object to mean julian date (MJD).

    Parameters
    ----------
    time_obj: datetime.datetime
        The datetime object.

    Returns
    -------
    float
        The mean julian date (MJD).
    """
    return (time_obj - datetime(1858, 11, 17)).total_seconds() / 86400


def primary_hdu(
        obs_info: dict,
        gnc_info: dict,
        filename_output=False):
    """
    Generate the primary hdu of the fits file.

    Parameters
    ----------
    obs_info: dict
        The parameters of the observation. See `save_fits` function.
    gnc_info: dict
        The gnc information of the observation.
    filename_output: bool (optional)
        If True, the folder and the filename will be returned.

    Returns
    -------
    fits.PrimaryHDU
        The primary hdu of the fits file.
    str, str
        The folder and filename of the fits file. Only returned if filename_output is True.

    Notes
    -----
    The gnc_info dict should contain the information of orbit and observation.
    these informations are used to genrated the hdu header. Refer to the data model for more information.
    """

    # camera_config, _ = load_camera_and_optics_config(obs_info['band'])

    obsid = obs_info['obsid']

    exp_start = obs_info['EXPSTART']
    exp_start = datetime.fromisoformat(exp_start)

    exp_end = obs_info['EXPEND']
    exp_end = datetime.fromisoformat(exp_end)

    filename = "CSST_CPIC"
    filename += "_" + which_focalplane(obs_info['band']).upper()
    filename += "_" + obsid_parser(obsid)
    filename += "_" + exp_start.strftime("%Y%m%d%H%M%S")
    filename += "_" + exp_end.strftime("%Y%m%d%H%M%S")
    filename += f"_{obsid}_X_L0_V01.fits"

    type_dir = 'CAL'
    if int(f'{obsid}'[1:3]) <= 10:
        type_dir = 'SCI'

    mjd_dir = f"{datetime_obj_to_mjd(exp_start):.0f}"
    folder = f"{type_dir}/{mjd_dir}"

    header = fits.Header()
    # General keywords
    header['SIMPLE'] = True
    header['BITPIX'] = 8
    header['NAXIS'] = 0
    header['EXTEND'] = True
    header['NEXTEND'] = 1  # + parameters['nframe']
    # header['GROUPS'] = False

    header['DATE'] = datetime.now().isoformat(timespec='seconds')
    heaer_filename = filename[:-4]
    if len(heaer_filename) > 68:
        heaer_filename = heaer_filename[:68]
    header['FILENAME'] = heaer_filename
    header['FILETYPE'] = obsid_parser(obsid)
    header['TELESCOP'] = 'CSST'
    header['INSTRUME'] = 'CPIC'
    header['RADECSYS'] = 'ICRS'
    header['EQUINOX'] = 2000.0
    header['FITSSWV'] = f'CPISM V{__version__}'
    header['COMMENT'] = ''

    cstar = {'ra': '0d', 'dec': '0d'}
    if obs_info['target'] != {}:
        cstar = obs_info['target']['cstar']

    radec = SkyCoord(cstar['ra'], cstar['dec'])
    target_name = radec.to_string('hmsdms')
    target_name = re.sub(R'[hdms\s]', '', target_name)
    header['OBJECT'] = cstar.get('name', target_name)
    header['TARGET'] = target_name
    header['OBSID'] = str(obsid)
    header['RA_OBJ'] = radec.ra.degree
    header['DEC_OBJ'] = radec.dec.degree

    # telescope information
    header['REFFRAME'] = 'CSSTGSC-1.0'
    header['DATE-OBS'] = exp_start.isoformat(timespec='seconds')
    header['SATESWV'] = '1'

    header['EXPSTART'] = datetime_obj_to_mjd(exp_start)
    cabstart = gnc_info.get('CABSTART', exp_start.isoformat(timespec='seconds'))
    cabstart = iso_time(cabstart)
    cabstart_mjd = datetime_obj_to_mjd(datetime.fromisoformat(cabstart))
    header['CABSTART'] = cabstart_mjd
    header['SUNANGL0'] = gnc_info.get('SUNANGL0', -1.0)
    header['MOONANG0'] = gnc_info.get('MOONANG0', -1.0)
    header['TEL_ALT0'] = gnc_info.get('TEL_ALT0', -1.0)
    header['POS_ANG0'] = gnc_info.get(
        'POS_ANG0', float(obs_info['rotation']))
    header['POSI0_X'] = gnc_info.get('POSI0_X', -1.0)
    header['POSI0_Y'] = gnc_info.get('POSI0_Y', -1.0)
    header['POSI0_Z'] = gnc_info.get('POSI0_Z', -1.0)
    header['VELO0_X'] = gnc_info.get('VELO0_X', -1.0)
    header['VELO0_Y'] = gnc_info.get('VELO0_Y', -1.0)
    header['VELO0_Z'] = gnc_info.get('VELO0_Z', -1.0)
    header['EULER0_1'] = gnc_info.get('EULER0_1', -1.0)
    header['EULER0_2'] = gnc_info.get('EULER0_2', -1.0)
    header['EULER0_3'] = gnc_info.get('EULER0_3', -1.0)
    header['RA_PNT0'] = gnc_info.get('RA_PNT0', header['RA_OBJ'])
    header['DEC_PNT0'] = gnc_info.get('DEC_PNT0', header['DEC_OBJ'])

    header['EXPEND'] = datetime_obj_to_mjd(exp_end)

    cabend = gnc_info.get('CABEND', exp_end.isoformat(timespec='seconds'))
    cabend = iso_time(cabend)
    cabend_mjd = datetime_obj_to_mjd(datetime.fromisoformat(cabend))
    header['CABEND'] = cabend_mjd
    header['SUNANGL1'] = gnc_info.get('SUNANGL1', header['SUNANGL0'])
    header['MOONANG1'] = gnc_info.get('MOONANG1', header['MOONANG0'])
    header['TEL_ALT1'] = gnc_info.get('TEL_ALT1', header['TEL_ALT0'])
    header['POS_ANG1'] = gnc_info.get('POS_ANG1', header['POS_ANG0'])
    header['POSI1_X'] = gnc_info.get('POSI1_X', header['POSI0_x'])
    header['POSI1_Y'] = gnc_info.get('POSI1_Y', header['POSI0_y'])
    header['POSI1_Z'] = gnc_info.get('POSI1_Z', header['POSI0_z'])
    header['VELO1_X'] = gnc_info.get('VELO1_X', header['VELO0_x'])
    header['VELO1_Y'] = gnc_info.get('VELO1_Y', header['VELO0_y'])
    header['VELO1_Z'] = gnc_info.get('VELO1_Z', header['VELO0_z'])
    header['EULER1_1'] = gnc_info.get('EULER1_1', header['EULER0_1'])
    header['EULER1_2'] = gnc_info.get('EULER1_2', header['EULER0_2'])
    header['EULER1_3'] = gnc_info.get('EULER1_3', header['EULER0_3'])
    header['RA_PNT1'] = gnc_info.get('RA_PNT1', header['RA_PNT0'])
    header['DEC_PNT1'] = gnc_info.get('DEC_PNT1', header['DEC_PNT0'])
    
    header['EPOCH'] = float(exp_start.year)
    header['EXPTIME'] = (exp_end - exp_start).total_seconds()
    
    header['CHECKSUM'] = '0000000000000000'
    header['DATASUM'] = '0000000000'

    check_and_update_fits_header(header)

    # other information
    hdu = fits.PrimaryHDU(header=header)
    hdu.add_checksum()

    if filename_output:
        return hdu, folder, filename
    else:
        return hdu


def frame_header(obs_info, index, primary_header, camera_dict):
    """
    Generate the header for a single frame.

    Parameters
    ----------
    obs_info : dict
        Dictionary of parameters. See `save_fits` function.
    index : int
        Frame index.
    primary_header : dict
        Primary header

    Returns
    -------
    astropy.io.fits.Header

    """

    header = fits.Header()
    # camera_config, optics_config = load_camera_and_optics_config(
    #     obs_info['band'])
    camera_config = camera_dict
    plszx = camera_config['plszx']
    plszy = camera_config['plszy']
    pscan1 = camera_config['pscan1']
    pscan2 = camera_config['pscan2']
    oscan1 = camera_config['oscan1']
    oscan2 = camera_config['oscan2']
    udark = camera_config['udark']
    bdark = camera_config['bdark']
    ldark = camera_config['ldark']
    rdark = camera_config['rdark']
    imgszx = plszx + pscan1 + oscan1 + ldark + rdark
    imgszy = plszy + pscan2 + oscan2 + udark + bdark

    header['XTENSION'] = 'IMAGE'
    header['BITPIX'] = 16
    header['NAXIS'] = 2
    header['NAXIS1'] = 1080
    header['NAXIS2'] = 1050
    header['PCOUNT'] = 0
    header['GCOUNT'] = 1
    header['BSCALE'] = 1
    header['BZERO'] = 32768
    header['EXTNAME'] = 'IMAGE'
    header['EXTVER'] = 1
    header['BUNIT'] = 'ADU'

    header['FILTER'] = obs_info['band']
    header['DETSN'] = '0'
    header['DETNAME'] = camera_config['detector_name']
    header['CHIPLAB'] = camera_config['ccd_label']
    header['DEWTEMP'] = float(camera_config['cooler_temp'])
    frame_info = obs_info['frame_info'][index]
    header['CHIPTEMP'] = float(frame_info['chiptemp'])
    header['DETSIZE'] = f"{imgszx} * {imgszy}"
    header['IMGINDEX'] = index + 1

    utc0_float = config['utc0_float']
    ra0 = primary_header.get('RA_PNT0', -1.0)
    dec0 = primary_header.get('DEC_PNT0', -1.0)
    pa0 = primary_header.get('POS_ANG0', -1.0)
    cab0 = primary_header.get('CABSTART', obs_info['frame_info'][0]['expt_start'])

    ra1 = primary_header.get('RA_PNT1', ra0)
    dec1 = primary_header.get('DEC_PNT1', dec0)
    pa1 = primary_header.get('POS_ANG1', pa0)
    cab1 = primary_header.get('CABEND', cab0)

    frame_stamp = frame_info['expt_start'] + utc0_float
    frame_mjd = datetime_obj_to_mjd(datetime.fromtimestamp(frame_stamp))

    cab0 = iso_time(cab0)
    cab1 = iso_time(cab1)
    cab0 = datetime_obj_to_mjd(datetime.fromisoformat(cab0))
    cab1 = datetime_obj_to_mjd(datetime.fromisoformat(cab1))

    ratio = (frame_mjd - cab0)/(cab1 - cab0)
    ra = ra0 + (ra1 - ra0) * ratio
    dec = dec0 + (dec1 - dec0) * ratio
    pa = pa0 + (pa1 - pa0) * ratio

    header['IMG_EXPT'] = datetime.fromtimestamp(frame_stamp).isoformat(timespec='seconds')
    header['IMG_CABT'] = header['IMG_EXPT']
    header['IMG_DUR'] = float(obs_info['expt'])

    header['IMG_PA'] = ra
    header['IMG_RA'] = dec
    header['IMG_DEC'] = pa

    header['DATASECT'] = f"{plszx} * {plszy}"
    header['PIXSCAL'] = frame_info['platescale']
    header['PIXSIZE'] = float(camera_config['pitch_size'])
    header['NCHAN'] = 1
    header['PSCAN1'] = pscan1
    header['PSCAN2'] = pscan2
    header['OSCAN1'] = oscan1
    header['OSCAN2'] = oscan2
    header['UDARK'] = udark
    header['BDARK'] = bdark
    header['LDARK'] = ldark
    header['RDARK'] = rdark

    # WCS
    cstar = {'ra': '0d', 'dec': '0d'}
    if obs_info['target'] != {}:
        cstar = obs_info['target']['cstar']

    radec = SkyCoord(cstar['ra'], cstar['dec'])
    shift = obs_info['shift']
    platescale = frame_info['platescale']
    rotation = np.radians(obs_info['rotation'])

    header['WCSAXES'] = 2
    header['CRPIX1'] = (plszx + 1)/2 + pscan1 + ldark + shift[0] / platescale
    header['CRPIX2'] = (plszy + 1)/2 + pscan2 + udark + shift[0] / platescale
    header['CRVAL1'] = radec.ra.degree
    header['CRVAL2'] = radec.dec.degree
    header['CTYPE1'] = 'RA---TAN'
    header['CTYPE2'] = 'DEC--TAN'
    header['CD1_1'] = np.cos(rotation)
    header['CD1_2'] = -np.sin(rotation)
    header['CD2_1'] = np.sin(rotation)
    header['CD2_2'] = np.cos(rotation)

    # Readout information
    header['EMGAIN'] = float(obs_info['emgain'])
    header['GAIN'] = float(camera_config['ph_per_adu'])
    header['DET_BIAS'] = float(camera_config['bias_level'])
    header['RON'] = float(camera_config['readout_noise'])
    header['READTIME'] = float(camera_config['readout_time'])
    header['ROSPEED'] = float(camera_config['readout_speed'])

    # CPIC information
    header['LS_STAT'] = 'OFF'
    header['IWA'] = frame_info['iwa']
    header['WFSINFO1'] = -1.0
    header['WFSINFO2'] = -1.0

    header['CHECKSUM'] = '0000000000000000'
    header['DATASUM'] = '0000000000'

    header = check_and_update_fits_header(header)

    return header


def save_fits_simple(images, obs_info, output_folder='./'):
    """
    Save the image to a fits file with a simple header to TMP directory.

    Parameters
    ----------
    images : numpy.ndarray
        Image array to be written.
    obs_info : dict
        Dictionary of observation informations. See `save_fits` function.

    Returns
    ----------
    Filename of the saved fits file.

    """
    target = obs_info['target']
    target_info = 'NO_TARGET'
    if 'cstar' in target.keys():
        target_info = ''
        target_info = f"S{target['cstar']['magnitude']:.1f}"
        target_info += f"_P{len(target.get('planets', '[]'))}"

    name = target_info
    if 'name' in target.keys():
        name = target['name']

    name = name.replace('/', '_')
    name = name.replace(',', '_')

    now = datetime.now()
    time = now.strftime("%Y%m%d%H%M%S")
    filename = f"{name}_{time}.fits"

    header = fits.Header()
    header['skybg'] = obs_info['skybg']
    header['name'] = name
    header['exptime'] = obs_info['expt']
    header['nframe'] = obs_info['nframe']
    header['band'] = obs_info['band']
    header['emgain'] = obs_info['emgain']
    header['obsid'] = obs_info['obsid']
    header['rotation'] = obs_info['rotation']
    shift = obs_info['shift']
    header['shift'] = f"x:{shift[0]},y:{shift[1]}"


    fullname = os.path.join(output_folder, filename)
    print(fullname)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
        log.debug(f"Output folder {output_folder} is created.")

    log.debug(f"save fits file to {fullname}")
    fits.writeto(fullname, images, overwrite=True, header=header)
    return fullname


def save_fits(images, obs_info, gnc_info, camera_dict={}, csst_format=True, output_folder='./'):
    """
    Save the image to a fits file.

    Parameters
    ----------
    images : numpy.ndarray
        Image array to be saved.
    obs_info : dict
        Dictionary of observation informations.
        Must contain the following keys

        - band: str
            - Band of the image.
        - expt: float
            - Exposure time of the each image.
        - nframe: int
            - Number of frames in the image.
        - emgain: int
            - EM gain of the camera.
        - obsid: str
            - Observation ID. Obsid must be 11 digits and start with 5 for CPIC. See pharse_obsid() for details.
        - rotation: float
            - Rotation angle of the image.
        - shift: list
            - Shift of the image.

    gnc_info : dict
        Dictionary of GNC information.
        Contains the keywords in the primary header. See primary_hdu() for details.

    csst_format : bool, optional
        Whether to save the fits file in CSST format, by default True.
    """

    if not csst_format:
        save_fits_simple(images, obs_info, output_folder=output_folder)
        return

    hdu_header, folder, filename = primary_hdu(obs_info, gnc_info, True)
    hdu_list = fits.HDUList([hdu_header])

    if len(images.shape) == 2:
        images = np.array([images])

    for index in range(images.shape[0]):
        header = frame_header(
            obs_info,
            index,
            hdu_header.header,
            camera_dict=camera_dict,
        )
        frame_hdu = fits.ImageHDU(images[index, :, :], header=header)
        frame_hdu.add_checksum()
        hdu_list.append(frame_hdu)

    folder = f"{output_folder}/{folder}"

    if not os.path.exists(folder):
        os.makedirs(folder)
        log.debug(f'make new folder {folder}')

    full_path = f"{folder}/{filename}"
    log.debug(f'save fits file: {full_path}')
    hdu_list.writeto(full_path, overwrite=True)
